import json

class FeatureFlagsState(object):
    """
    A snapshot of the state of all feature flags with regard to a specific user, generated by
    calling the client's all_flags_state method. Serializing this object to JSON, using the
    to_json_dict method or jsonpickle, will produce the appropriate data structure for
    bootstrapping the LaunchDarkly JavaScript client.
    """
    def __init__(self, valid):
        self.__flag_values = {}
        self.__flag_metadata = {}
        self.__valid = valid

    def add_flag(self, flag, value, variation, reason):
        """Used internally to build the state map."""
        key = flag['key']
        self.__flag_values[key] = value
        meta = { 'version': flag.get('version'), 'trackEvents': flag.get('trackEvents') }
        if variation is not None:
            meta['variation'] = variation
        if reason is not None:
            meta['reason'] = reason
        if flag.get('debugEventsUntilDate') is not None:
            meta['debugEventsUntilDate'] = flag.get('debugEventsUntilDate')
        self.__flag_metadata[key] = meta
    
    @property
    def valid(self):
        """True if this object contains a valid snapshot of feature flag state, or False if the
        state could not be computed (for instance, because the client was offline or there was no user).
        """
        return self.__valid
    
    def get_flag_value(self, key):
        """Returns the value of an individual feature flag at the time the state was recorded.
        :param string key: the feature flag key
        :return: the flag's value; None if the flag returned the default value, or if there was no such flag
        """
        return self.__flag_values.get(key)
    
    def get_flag_reason(self, key):
        """Returns the evaluation reason for an individual feature flag at the time the state was recorded.
        :param string key: the feature flag key
        :return: a dictionary describing the reason; None if reasons were not recorded, or if there was no
          such flag
        """
        meta = self.__flag_metadata.get(key)
        return None if meta is None else meta.get('reason')
    
    def to_values_map(self):
        """Returns a dictionary of flag keys to flag values. If the flag would have evaluated to the
        default value, its value will be None.

        Do not use this method if you are passing data to the front end to "bootstrap" the JavaScript client.
        Instead, use to_json_dict.
        """
        return self.__flag_values

    def to_json_dict(self):
        """Returns a dictionary suitable for passing as JSON, in the format used by the LaunchDarkly
        JavaScript SDK. Use this method if you are passing data to the front end in order to
        "bootstrap" the JavaScript client.
        """
        ret = self.__flag_values.copy()
        ret['$flagsState'] = self.__flag_metadata
        ret['$valid'] = self.__valid
        return ret
    
    def to_json_string(self):
        """Same as to_json_dict, but serializes the JSON structure into a string.
        """
        return json.dumps(self.to_json_dict())

    def __getstate__(self):
        """Equivalent to to_json_dict() - used if you are serializing the object with jsonpickle.
        """
        return self.to_json_dict()
